"""Piece together the actual VAE CNN model for tokun."""

import functools

import keras
import tensorflow as tf

import mlable.layers.shaping

# ENCODER #####################################################################

@keras.saving.register_keras_serializable(package='models')
class Encoder(tf.keras.models.Model):
    def __init__(
        self,
        embed_dim: int,
        input_dim: int=256,
        **kwargs
    ) -> None:
        # init
        super(Encoder, self).__init__(**kwargs)
        # config
        self._config = {
            'input_dim': input_dim,
            'embed_dim': embed_dim,}
        # layers
        self._layers = []

    def build(self, input_shape: tuple) -> None:
        __shape = tuple(input_shape)
        # weights
        __q, __r = tf.linalg.qr(tf.random.normal((self._config['input_dim'], self._config['embed_dim'])))
        __e = __q * tf.sign(tf.linalg.diag_part(tf.expand_dims(__r, axis=0)))
        # init
        self._layers = [
            # (B, ..., T) => (B, ..., T, E)
            tf.keras.layers.Embedding(
                input_dim=self._config['input_dim'],
                output_dim=self._config['embed_dim'],
                embeddings_initializer=tf.keras.initializers.Constant(__e),
                name='encoder-0-embed'),
            # (B, ..., T, E) => (B, ..., T * E)
            mlable.layers.shaping.Merge(
                axis=-1,
                right=False),]
        # build
        for __l in self._layers:
            __l.build(__shape)
            __shape = __l.compute_output_shape(__shape)
        # register
        self.built = True

    def compute_embedding_kl_loss(self, beta: float=1.0) -> tf.Tensor:
        __e = self._layers[0].embeddings
        # mean (E,)
        __m = tf.reduce_mean(__e, axis=0)
        # variance (E,)
        __v = tf.math.reduce_variance(__e, axis=0)
        # KL divergence ()
        return beta * 0.5 * tf.reduce_sum(__v + tf.square(__m) - 1. - tf.math.log(__v + 1e-8))

    def compute_output_shape(self, input_shape: tuple) -> tuple:
        return functools.reduce(lambda __s, __l: __l.compute_output_shape(__s), self._layers, input_shape)

    def call(self, inputs: tf.Tensor, training: bool=False, **kwargs) -> tf.Tensor:
        # KL divergence of the embeddings to N(0, I)
        if training:
            self.add_loss(tf.cast(self.compute_embedding_kl_loss(beta=1.0), dtype=tf.float32))
        # composite embeddings, token by token (B, ..., T) => (B, ..., T * E)
        return functools.reduce(lambda __x, __l: __l(__x, **kwargs), self._layers, inputs)

    def get_config(self) -> dict:
        __config = super(Encoder, self).get_config()
        __config.update(self._config)
        return __config

    @classmethod
    def from_config(cls, config) -> tf.keras.layers.Layer:
        return cls(**config)

# DECODER #####################################################################

@keras.saving.register_keras_serializable(package='models')
class Decoder(tf.keras.models.Model):
    def __init__(
        self,
        token_dim: int,
        binary_dim: int=8, # return byte predictions
        **kwargs
    ) -> None:
        # init
        super(Decoder, self).__init__(**kwargs)
        # config
        self._config = {
            'token_dim': token_dim,
            'binary_dim': binary_dim,}
        # layers
        self._layers = []

    def build(self, input_shape: tuple) -> tuple:
        __shape = tuple(input_shape)
        # init
        self._layers = [
            # (B, ..., L) => (B, ..., T * O)
            tf.keras.layers.Dense(
                units=self._config['token_dim'] * self._config['binary_dim'],
                use_bias=True,
                activation=None,
                kernel_initializer='zeros',
                bias_initializer='zeros',
                name='projection')]
        # build
        for __l in self._layers:
            __l.build(__shape)
            __shape = __l.compute_output_shape(__shape)
        # register
        self.built = True

    def compute_output_shape(self, input_shape: tuple) -> tuple:
        return functools.reduce(lambda __s, __l: __l.compute_output_shape(__s), self._layers, input_shape)

    def call(self, inputs: tf.Tensor, logits: bool=True, training: bool=False, **kwargs) -> tf.Tensor:
        __outputs = functools.reduce(lambda __x, __l: __l(__x, **kwargs), self._layers, inputs)
        # raw logits or probabilities
        return __outputs if logits else tf.sigmoid(__outputs)

    def get_config(self) -> dict:
        __config = super(Decoder, self).get_config()
        __config.update(self._config)
        return __config

    @classmethod
    def from_config(cls, config) -> tf.keras.layers.Layer:
        return cls(**config)

# VAE #########################################################################

@keras.saving.register_keras_serializable(package='models')
class QuantizedAutoEncoder(tf.keras.models.Model):
    def __init__(
        self,
        token_dim: int,
        embed_dim: int,
        input_dim: int=256, # byte inputs
        binary_dim: int=8, # byte predictions
        **kwargs
    ) -> None:
        # init
        super(QuantizedAutoEncoder, self).__init__(**kwargs)
        # config
        self._config = {
            'token_dim': token_dim,
            'input_dim': input_dim,
            'embed_dim': embed_dim,
            'binary_dim': binary_dim,}
        # layers
        self._encoder = None
        self._decoder = None

    def build(self, input_shape: tuple) -> None:
        __shape = tuple(input_shape)
        # init
        self._encoder = Encoder(
            input_dim=self._config['input_dim'],
            embed_dim=self._config['embed_dim'])
        self._decoder = Decoder(
            token_dim=self._config['token_dim'],
            binary_dim=self._config['binary_dim'],)
        # build
        self._encoder.build(__shape)
        __shape = self._encoder.compute_output_shape(__shape)
        self._decoder.build(__shape)
        __shape = self._decoder.compute_output_shape(__shape)
        # register
        self.built = True

    def compute_latent_shape(self, input_shape: tuple) -> tuple:
        return self._encoder.compute_output_shape(input_shape)

    def compute_output_shape(self, input_shape: tuple) -> tuple:
        return self._decoder.compute_output_shape(self._encoder.compute_output_shape(input_shape))

    def encode(self, inputs: tf.Tensor, training: bool=False, **kwargs) -> tf.Tensor:
        return self._encoder(inputs, training=training, **kwargs)

    def decode(self, inputs: tf.Tensor, logits: bool=True, training: bool=False, **kwargs) -> tf.Tensor:
        return self._decoder(inputs, logits=logits, training=training, **kwargs)

    def call(self, inputs: tf.Tensor, logits: bool=True, training: bool=False, **kwargs) -> tf.Tensor:
        # composite embeddings (B, ..., T) => (B, ..., T * E)
        __outputs = self._encoder(inputs, training=training, **kwargs)
        # binary encoding of each token (B, ..., T * E) => (B, ..., T * U)
        return self._decoder(__outputs, logits=logits, training=training, **kwargs)

    def get_config(self) -> dict:
        __config = super(QuantizedAutoEncoder, self).get_config()
        __config.update(self._config)
        return __config

    @classmethod
    def from_config(cls, config) -> tf.keras.layers.Layer:
        return cls(**config)



"""Piece together the actual VAE CNN model for tokun."""

import functools

import tensorflow as tf

import mlable.blocks.convolution.resnet
import mlable.blocks.convolution.unet
import mlable.layers.embedding
import mlable.layers.shaping
import mlable.models.autoencoder

# CONSTANTS ###################################################################

DROPOUT = 0.0
EPSILON = 1e-6

# ENCODER #####################################################################

@tf.keras.utils.register_keras_serializable(package='models')
class Encoder(tf.keras.models.Model):
    def __init__(
        self,
        block_cfg: iter,
        embed_dim: int,
        input_dim: int=256,
        dropout_rate: float=DROPOUT,
        epsilon_rate: float=EPSILON,
        **kwargs
    ) -> None:
        super(Encoder, self).__init__(**kwargs)
        # save for IO serialization
        self._config = {
            'block_cfg': block_cfg,
            'input_dim': max(1, input_dim),
            'embed_dim': max(1, embed_dim),
            'dropout_rate': max(0.0, dropout_rate),
            'epsilon_rate': max(1e-8, epsilon_rate),}
        # layers
        self._embed_byte = None
        self._merge_bytes = None
        self._embed_height = None
        self._embed_width = None
        self._unet_blocks = []

    def build(self, input_shape: tuple) -> None:
        __shape = tuple(input_shape)
        # embed (B, H, W, C) => (B, H, W, C, E)
        self._embed_byte = tf.keras.layers.Embedding(
            name='encoder-0-embed-bytes',
            **self.get_embed_byte_config())
        # merge (B, H, W, C, E) => (B, H, W, CE)
        self._merge_bytes = mlable.layers.shaping.Merge(
            name='encoder-1-merge-embeds',
            **self.get_merge_config())
        # embed height (B, H, W, C) => (B, H, W, CE)
        self._embed_height = mlable.layers.embedding.PositionalEmbedding(
            name='encoder-2-embed-height',
            **self.get_embed_height_config())
        # embed width (B, H, W, C) => (B, H, W, CE)
        self._embed_width = mlable.layers.embedding.PositionalEmbedding(
            name='encoder-3-embed-width',
            **self.get_embed_width_config())
        # compress (B, H/2^i, W/2^i, Li) => (B, H/2^i+1, W/2^i+1, Li+1)
        self._unet_blocks = [
            mlable.blocks.convolution.unet.UnetBlock(
                name=f'encoder-{__i + 4}-unet-down',
                **__c)
            for __i, __c in enumerate(self.get_unet_configs())]
        # build
        for __b in self.get_all_blocks():
            __b.build(__shape)
            __shape = __b.compute_output_shape(__shape)
        # register
        self.built = True

    def call(self, inputs: tf.Tensor, training: bool=False, **kwargs) -> tf.Tensor:
        return functools.reduce(lambda __x, __b: __b(__x, training=training, **kwargs), self.get_all_blocks(), inputs)

    def compute_output_shape(self, input_shape: tuple) -> tuple:
        return functools.reduce(lambda __s, __b: __b.compute_output_shape(__s), self.get_all_blocks(), input_shape)

    def get_all_blocks(self) -> list:
        return [self._embed_byte, self._merge_bytes, self._embed_height, self._embed_width] + self._unet_blocks

    def get_config(self) -> dict:
        __config = super(Encoder, self).get_config()
        __config.update(self._config)
        return __config

    def get_embed_byte_config(self) -> dict:
        return {'input_dim': self._config['input_dim'], 'output_dim': self._config['embed_dim'],}

    def get_embed_height_config(self) -> dict:
        return {'sequence_axis': 1, 'feature_axis': -1,}

    def get_embed_width_config(self) -> dict:
        return {'sequence_axis': 2, 'feature_axis': -1,}

    def get_merge_config(self) -> dict:
        return {'axis': -1, 'right': False,}

    def get_unet_configs(self) -> dict:
        return [
            {
                'channel_dim': self._config['block_cfg'][0]['channel_dim'],
                'group_dim': None,
                'head_dim': None,
                'head_num': None,
                'layer_num': 1,
                'add_attention': False,
                'add_downsampling': False,
                'add_upsampling': False,
                'dropout_rate': self._config['dropout_rate'],
                'epsilon_rate': self._config['epsilon_rate'],}] + [
            {
                'channel_dim': __c['channel_dim'],
                'group_dim': __c.get('group_dim', None),
                'head_dim': __c.get('head_dim', None),
                'head_num': __c.get('head_num', None),
                'layer_num': __c.get('layer_num', 2),
                'add_attention': __c.get('add_attention', False),
                'add_downsampling': __c.get('add_downsampling', False),
                'add_upsampling': __c.get('add_upsampling', False),
                'dropout_rate': self._config['dropout_rate'],
                'epsilon_rate': self._config['epsilon_rate'],}
            for __c in self._config['block_cfg']] + [
            {
                'channel_dim': 2 * self._config['block_cfg'][-1]['channel_dim'],
                'group_dim': None,
                'head_dim': None,
                'head_num': None,
                'layer_num': 1,
                'add_attention': False,
                'add_downsampling': False,
                'add_upsampling': False,
                'dropout_rate': self._config['dropout_rate'],
                'epsilon_rate': self._config['epsilon_rate'],}]

    @classmethod
    def from_config(cls, config) -> tf.keras.layers.Layer:
        return cls(**config)

# DECODER #####################################################################

@tf.keras.utils.register_keras_serializable(package='models')
class Decoder(tf.keras.models.Model):
    def __init__(
        self,
        block_cfg: iter,
        output_dim: int,
        dropout_rate: float=DROPOUT,
        epsilon_rate: float=EPSILON,
        **kwargs
    ) -> None:
        super(Decoder, self).__init__(**kwargs)
        # save for IO serialization
        self._config = {
            'block_cfg': block_cfg,
            'output_dim': max(1, output_dim),
            'dropout_rate': max(0.0, dropout_rate),
            'epsilon_rate': max(1e-8, epsilon_rate),}
        # blocks
        self._unet_blocks = []

    def build(self, input_shape: tuple) -> None:
        __shape = tuple(input_shape)
        # decompress (B, H/2^i+1, W/2^i+1, Li+1) => (B, H/2^i, W/2^i, Li)
        self._unet_blocks = [
            mlable.blocks.convolution.unet.UnetBlock(
                name=f'decoder-{__i}-unet-up',
                **__c)
            for __i, __c in enumerate(self.get_unet_configs())]
        # build
        for __b in self._unet_blocks:
            __b.build(__shape)
            __shape = __b.compute_output_shape(__shape)
        # register
        self.built = True

    def compute_output_shape(self, input_shape: tuple) -> tuple:
        return functools.reduce(lambda __s, __b: __b.compute_output_shape(__s), self._unet_blocks, input_shape)

    def call(self, inputs: tf.Tensor, training: bool=False, **kwargs) -> tf.Tensor:
        return functools.reduce(lambda __x, __b: __b(__x, training=training, **kwargs), self._unet_blocks, inputs)

    def get_config(self) -> dict:
        __config = super(Decoder, self).get_config()
        __config.update(self._config)
        return __config

    def get_unet_configs(self) -> dict:
        return [
            {
                'channel_dim': __c['channel_dim'],
                'group_dim': __c.get('group_dim', None),
                'head_dim': __c.get('head_dim', None),
                'head_num': __c.get('head_num', None),
                'layer_num': __c.get('layer_num', 2),
                'add_attention': __c.get('add_attention', False),
                'add_downsampling': __c.get('add_downsampling', False),
                'add_upsampling': __c.get('add_upsampling', False),
                'dropout_rate': self._config['dropout_rate'],
                'epsilon_rate': self._config['epsilon_rate'],}
            for __c in self._config['block_cfg']] + [
            {
                'channel_dim': self._config['output_dim'],
                'group_dim': None,
                'head_dim': None,
                'head_num': None,
                'layer_num': 2,
                'add_attention': False,
                'add_downsampling': False,
                'add_upsampling': False,
                'dropout_rate': self._config['dropout_rate'],
                'epsilon_rate': self._config['epsilon_rate'],}]

    @classmethod
    def from_config(cls, config) -> tf.keras.layers.Layer:
        return cls(**config)

# VAE #########################################################################

@tf.keras.utils.register_keras_serializable(package='models')
class KlAutoEncoder(mlable.models.autoencoder.VaeModel):
    def __init__(
        self,
        block_cfg: iter,
        embed_dim: int,
        output_dim: int,
        input_dim: int=256,
        step_min: int=0,
        step_max: int=2 ** 12,
        beta_min: float=0.0,
        beta_max: float=1.0,
        dropout_rate: float=DROPOUT,
        epsilon_rate: float=EPSILON,
        **kwargs
    ) -> None:
        super(KlAutoEncoder, self).__init__(step_min=step_min, step_max=step_max, beta_min=beta_min, beta_max=beta_max, **kwargs)
        # save for IO serialization
        self._config.update({
            'block_cfg': block_cfg,
            'input_dim': max(1, input_dim),
            'embed_dim': max(1, embed_dim),
            'output_dim': max(1, output_dim),
            'dropout_rate': max(0.0, dropout_rate),
            'epsilon_rate': max(1e-8, epsilon_rate),})
        # layers
        self._encoder = None
        self._decoder = None

    def build(self, input_shape: tuple) -> None:
        # init
        self._encoder = Encoder(**self.get_encoder_config())
        self._decoder = Decoder(**self.get_decoder_config())
        # build
        __shape = tuple(input_shape)
        self._encoder.build(__shape)
        __shape = self.compute_latent_shape(__shape)
        self._decoder.build(__shape)
        # register
        self.built = True

    def encode(self, inputs: tf.Tensor, training: bool=False, **kwargs) -> tuple:
        __outputs = self._encoder(inputs, training=training, **kwargs)
        # split in 2: mean + logvar
        return tuple(tf.split(__outputs, num=2, num_or_size_splits=2, axis=-1))

    def decode(self, inputs: tf.Tensor, training: bool=False, **kwargs) -> tf.Tensor:
        return self._decoder(inputs, training=training, **kwargs)

    def compute_latent_shape(self, input_shape: tuple) -> tuple:
        __shape = self._encoder.compute_output_shape(input_shape)
        # split in 2, because the encoder returns both mean and logvar
        return tuple(__shape[:-1]) + (__shape[-1] // 2,)

    def compute_output_shape(self, input_shape: tuple) -> tuple:
        return self._decoder.compute_output_shape(self.compute_latent_shape(input_shape))

    def get_config(self) -> dict:
        __config = super(KlAutoEncoder, self).get_config()
        __config.update(self._config)
        return __config

    def get_encoder_config(self) -> dict:
        return {
            'block_cfg': [
                {
                    'add_downsampling' if ('sampling' in __k) else __k: __v
                    for __k, __v in __c.items()}
                for __c in self._config['block_cfg']],
            **{__k: self._config[__k] for __k in ['input_dim', 'embed_dim', 'dropout_rate', 'epsilon_rate']}}

    def get_decoder_config(self) -> dict:
        return {
            'block_cfg': [
                {
                    'add_upsampling' if ('sampling' in __k) else __k: __v
                    for __k, __v in __c.items()}
                for __c in reversed(self._config['block_cfg'])],
            **{__k: self._config[__k] for __k in ['output_dim', 'dropout_rate', 'epsilon_rate']}}

    @classmethod
    def from_config(cls, config) -> tf.keras.layers.Layer:
        return cls(**config)



"""Encoding and decoding blocks of tokun."""

import functools

import keras
import tensorflow as tf

import mlable.layers.shaping
import mlable.layers.transformer

# ENCODING BLOCKS #############################################################

@keras.saving.register_keras_serializable(package='blocks')
class TokenizeBlock(tf.keras.layers.Layer):
    def __init__(
        self,
        sequence_axis: int=1,
        feature_axis: int=-1,
        token_dim: int=4,
        latent_dim: int=4,
        activation: str='gelu',
        epsilon: float=1e-6,
        **kwargs
    ) -> None:
        super(TokenizeBlock, self).__init__(**kwargs)
        # config
        self._config = {
            'sequence_axis': sequence_axis,
            'feature_axis': feature_axis,
            'token_dim': token_dim,
            'latent_dim': latent_dim,
            'activation': activation,
            'epsilon': epsilon,}
        # layers
        self._layers = []

    def build(self, input_shape: tuple) -> None:
        __shape = tuple(input_shape)
        # init
        self._layers = [
            # normalize each token unit independently
            tf.keras.layers.LayerNormalization(
                axis=self._config['feature_axis'],
                epsilon=self._config['epsilon'],
                center=True,
                scale=True,
                name='normalization'),
            # (B, S * G, E) => (B, S, G * E)
            mlable.layers.shaping.Divide(
                axis=self._config['sequence_axis'],
                factor=self._config['token_dim'],
                insert=False,
                right=True,
                name='reshaping'),
            # (B, S, G * E) => (B, S, L), typically L = E
            tf.keras.layers.Dense(
                units=self._config['latent_dim'],
                activation=self._config['activation'],
                use_bias=True,
                kernel_initializer='glorot_uniform',
                bias_initializer='zeros',
                name='compression'),]
        # build
        for __l in self._layers:
            __l.build(__shape)
            __shape = __l.compute_output_shape(__shape)
        # register
        self.built = True

    def compute_output_shape(self, input_shape: tuple) -> tuple:
        return functools.reduce(lambda __s, __l: __l.compute_output_shape(__s), self._layers, input_shape)

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        return functools.reduce(lambda __t, __l: __l(__t), self._layers, inputs)

    def get_config(self) -> dict:
        __config = super(TokenizeBlock, self).get_config()
        __config.update(self._config)
        return __config

    @classmethod
    def from_config(cls, config) -> tf.keras.layers.Layer:
        return cls(**config)

# DECODING BLOCKS #############################################################

@keras.saving.register_keras_serializable(package='blocks')
class DetokenizeBlock(tf.keras.layers.Layer):
    def __init__(
        self,
        sequence_axis: int=1,
        feature_axis: int=-1,
        token_dim: int=4,
        latent_dim: int=4,
        activation: str='gelu',
        epsilon: float=1e-6,
        **kwargs
    ) -> None:
        super(DetokenizeBlock, self).__init__(**kwargs)
        # config
        self._config = {
            'sequence_axis': sequence_axis,
            'feature_axis': feature_axis,
            'token_dim': token_dim,
            'latent_dim': latent_dim,
            'activation': activation,
            'epsilon': epsilon,}
        # layers
        self._layers = []

    def build(self, input_shape: tuple) -> None:
        __shape = tuple(input_shape)
        # init
        self._layers = [
            # (B, S, L) => (B, S, G * E), typically L = E
            tf.keras.layers.Dense(
                units=self._config['token_dim'] * self._config['latent_dim'],
                activation=self._config['activation'],
                use_bias=True,
                kernel_initializer='glorot_uniform',
                bias_initializer='zeros',
                name='decompression'),
            # (B, S, G * E) => (B, S * G, E)
            mlable.layers.shaping.Divide(
                axis=self._config['feature_axis'],
                factor=self._config['token_dim'],
                insert=False,
                right=False,
                name='reshaping'),
            # normalize each token unit independently
            tf.keras.layers.LayerNormalization(
                axis=self._config['feature_axis'],
                epsilon=self._config['epsilon'],
                center=True,
                scale=True,
                name='normalization'),]
        # build
        for __l in self._layers:
            __l.build(__shape)
            __shape = __l.compute_output_shape(__shape)
        # register
        self.built = True

    def compute_output_shape(self, input_shape: tuple) -> tuple:
        return functools.reduce(lambda __s, __l: __l.compute_output_shape(__s), self._layers, input_shape)

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        return functools.reduce(lambda __t, __l: __l(__t), self._layers, inputs)

    def get_config(self) -> dict:
        __config = super(DetokenizeBlock, self).get_config()
        __config.update(self._config)
        return __config

    @classmethod
    def from_config(cls, config) -> tf.keras.layers.Layer:
        return cls(**config)

# HEAD ########################################################################

@keras.saving.register_keras_serializable(package='blocks')
class HeadBlock(tf.keras.layers.Layer):
    def __init__(
        self,
        output_dim: int=8,
        **kwargs
    ) -> None:
        super(HeadBlock, self).__init__(**kwargs)
        # config
        self._config = {'output_dim': output_dim,}
        # layers
        self._dense = None # (..., G, E) => (..., G, U), typically U = E

    def build(self, input_shape: tuple) -> None:
        __shape = tuple(input_shape)
        # init
        self._dense = tf.keras.layers.Dense(
            units=self._config['output_dim'],
            use_bias=True,
            activation='linear',
            kernel_initializer='zeros',
            bias_initializer='zeros',
            name='projection')
        # build
        self._dense.build(__shape)
        # register
        self.built = True

    def compute_output_shape(self, input_shape: tuple) -> tuple:
        return tuple(input_shape[:-1]) + (self._config['output_dim'],)

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        return self._dense(inputs)

    def get_config(self) -> dict:
        __config = super(HeadBlock, self).get_config()
        __config.update(self._config)
        return __config

    @classmethod
    def from_config(cls, config) -> tf.keras.layers.Layer:
        return cls(**config)



import copy
import json

import transformers

# UTF-32-BE TOKENIZATION ######################################################

class ByteTokenizer(transformers.PreTrainedTokenizer):
    """
    Special tokenizer that encodes text as a sequence of byte indexes.

    It behaves like a regular tokenizer, with a vocabulary of 256 (1 byte, 8 bits).
    Most of the logic is inherited from `PreTrainedTokenizer`.

    The special tokens are set to obsolete / unused Unicode codepoints.
    These codepoints have legacy meanings that can be leveraged.

    UTF-8 encoding is the most compact encoding scheme, but the number of
    bytes to encode a given character can very between 1 and 4.

    UTF-32-BE encoding is very sparse since it systematically represents
    characters with 4 bytes, most of which are 0.
    However, the fixed size allows to patch the input without overlap.

    The vocabulary is set for compatibility with the parent classes.
    In UTF-32-BE "a" is '\u0061' or `[0, 0, 0, 97]`.
    The vocabulary is used to convert the indexes back and forth but the
    actual character represented comes from the combination of indexes.

    Args:
        encoding (`str, defaults to 'utf-8'):
            The text to integer mapping used.
            'utf-8' is the popular choice, but for ML purposes 'utf-32-be' is recommanded.
        bos_token (`str` or `tokenizers.AddedToken`, *optional*, defaults to '\u0002'):
            A special token representing the beginning of a sentence.
            Defaults to '\u0002', which the unicode codepoint for "start of text".
        eos_token (`str` or `tokenizers.AddedToken`, *optional*, defaults to '\u0003'):
            A special token representing the end of a sentence.
            Defaults to '\u0003', which the unicode codepoint for "end of text".
        unk_token (`str` or `tokenizers.AddedToken`, *optional*, defaults to '\u0000'):
            A special token representing an out-of-vocabulary token.
            Defaults to '\u0000', which the unicode codepoint for "null".
        sep_token (`str` or `tokenizers.AddedToken`, *optional*, defaults to '\u001e'):
            A special token separating two different sentences in the same input.
            Defaults to '\u001e', which the unicode codepoint for "record separator".
        pad_token (`str` or `tokenizers.AddedToken`, *optional*, defaults to '\u0080'):
            A special token used to make arrays of tokens the same size for batching purpose.
            Will then be ignored by attention mechanisms or loss computation.
            Defaults to '\u0080', which the unicode codepoint for "padding character".
        cls_token (`str` or `tokenizers.AddedToken`, *optional*, defaults to '\u001d'):
            A special token representing the class of the input (used by BERT for instance).
            Defaults to '\u001d', which the unicode codepoint for "group separator".
        mask_token (`str` or `tokenizers.AddedToken`, *optional*, defaults to '\u001a'):
            A special token representing a masked token.
            Defaults to '\u001a', which the unicode codepoint for "substitute".
    """

    def __init__(
        self,
        encoding: str='utf-8', # popular, but utf-32-be is recommanded
        bos_token: str='\u0002', # unicode "start of text"
        eos_token: str='\u0003', # unicode "end of text"
        unk_token: str='\u0000', # unicode "null"
        sep_token: str='\u001e', # unicode "record separator"
        pad_token: str='\u0080', # unicode "padding character"
        cls_token: str='\u001d', # unicode "group separator"
        mask_token: str='\u001a', # unicode "substitute"
        **kwargs,
    ) -> None:
        __kwargs = copy.deepcopy(kwargs)
        # properties
        self._encoding = encoding
        # enforce defaults
        __kwargs['additional_special_tokens'] = None # use the built-in special characters from Unicode
        __kwargs['split_special_tokens'] = True # in UTF-32, split the special codepoints into 4 bytes too
        # init
        super(ByteTokenizer, self).__init__(
            bos_token=bos_token,
            eos_token=eos_token,
            unk_token=unk_token,
            sep_token=sep_token,
            pad_token=pad_token,
            cls_token=cls_token,
            mask_token=mask_token,
            **__kwargs)

    def _tokenize(self, text: str, **kwargs) -> list:
        return list(chr(__b) for __b in text.encode(self._encoding))

    def _convert_token_to_id(self, token: str) -> int:
        return ord(token)

    def _convert_id_to_token(self, index: int) -> str:
        return chr(index)

    def convert_tokens_to_string(self, tokens: iter) -> str:
        return bytes(ord(__c) for __c in tokens).decode(self._encoding, errors="ignore")

    @property
    def vocab_size(self) -> int:
        return 256

    def get_vocab(self) -> dict: # for compatibility
        return {chr(__i): __i for __i in range(256)}

    def save_vocabulary(self, save_directory: str, **kwargs) -> tuple: # for compatibility
        __prefix = kwargs.get('filename_prefix', '')
        __path = "{}/{}vocab.json".format(save_directory, __prefix or '')
        with open(__path, "w") as __file:
            json.dump(self.get_vocab(), __file)
        return (__path,)

    def build_inputs_with_special_tokens(self, token_ids_0: list, token_ids_1: list=None) -> list:
        __bos_ids = [ord(__t) for __t in self._tokenize(self.bos_token)]
        __eos_ids = [ord(__t) for __t in self._tokenize(self.eos_token)]
        __cls_ids = [ord(__t) for __t in self._tokenize(self.cls_token)]
        __token_ids = token_ids_0 + __cls_ids + token_ids_1 if token_ids_1 else token_ids_0
        return __bos_ids + __token_ids + __eos_ids

    def get_special_tokens_mask(self, token_ids_0: list, token_ids_1: list=None, already_has_special_tokens: bool=False) -> list:
        if already_has_special_tokens:
            return super().get_special_tokens_mask(
                token_ids_0=token_ids_0,
                token_ids_1=token_ids_1,
                already_has_special_tokens=already_has_special_tokens)
        # mask matching `build_inputs_with_special_tokens`
        if token_ids_1:
            return [1] + ([0] * len(token_ids_0)) + [1] + ([0] * len(token_ids_1)) + [1]
        # default
        return [1] + ([0] * len(token_ids_0)) + [1]

    def create_token_type_ids_from_sequences(self, token_ids_0: list, token_ids_1: list=None) -> list:
        if token_ids_1:
            return (len(token_ids_0 + token_ids_1) + 3) * [0] # count the special tokens
        return (len(token_ids_0) + 2) * [0]
